"""
Iteration printers for dynamic period-by-period systems
"""


#[
from __future__ import annotations

from ..evaluators import printers as _printers
#]


class IterPrinter(_printers.BasicIterPrinter, ):
    pass

